
// Copyright (C) 2008 Indrek Mandre <indrek@mare.ee>
// http://www.mare.ee/indrek/timer/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#include "xavr/xavr.h"

typedef XAVR::Pins<XAVR::PortD::Out1> BEEP;

typedef XAVR::Display7<XAVR::PortC::Out1, XAVR::PortC::Out2,
    XAVR::PortB::Out5, XAVR::PortB::Out4, XAVR::PortB::Out3,
    XAVR::PortB::Out1, XAVR::PortB::Out2> D0;
typedef XAVR::Display7<XAVR::PortC::Out5, XAVR::PortC::Out0,
    XAVR::PortD::Out5, XAVR::PortD::Out6, XAVR::PortD::Out7,
    XAVR::PortC::Out4, XAVR::PortC::Out3> D1;

typedef XAVR::PortB::In0 B_RESET;
typedef XAVR::PortD::In2 B_UP;
typedef XAVR::PortD::In3 B_DOWN;

static void error(uint8_t n = 0)
{
  while (1)
    {
      D0::Set(n/10);
      D1::Set(n-(n/10)*10);
      for ( uint32_t i = 0; i < 1000; i++ )
        __asm__ __volatile__ ("nop");
      D0::Clear();
      D1::Clear();
      for ( uint32_t i = 0; i < 1000; i++ )
        __asm__ __volatile__ ("nop");
    }
}

static volatile uint16_t global_time = 0;
ISR(TIMER1_OVF_vect)
{
  global_time++;
}

// get global time since start in seconds
static uint32_t time()
{
  bool di = SREG & _BV(7);
  if ( di ) cli();
  // manually process the interrupt to prevent the clock from going backwards
  // in case of rollover
  if ( TIFR1 & _BV(TOV1) )
    {
      global_time++;
      TIFR1 &= ~_BV(TOV1);
    }
  uint32_t ret = (uint32_t)global_time * 512 + (uint32_t)(TCNT1 / 128);
  if ( di ) sei();
  return ret;
}

// get global time in milliseconds
static uint32_t timeMS()
{
  bool di = SREG & _BV(7);
  if ( di ) cli();
  if ( TIFR1 & _BV(TOV1) )
    {
      global_time++;
      TIFR1 &= ~_BV(TOV1);
    }
  uint32_t ret = (uint32_t)global_time * 512 * 1000 + (uint32_t)TCNT1 * 1000 / 128;
  if ( di ) sei();
  return ret;
}

#define KEY_PRESS_DELAY 25
#define KEY_HOLD_DELAY 300
#define KEY_REPEAT_DELAY 75

// a typing input key class. takes care of bouncing and repeats
struct InputKeyBase
{
protected:
  uint32_t time_off;
  uint32_t time_on;
  uint8_t key_count;
  bool flow;
  volatile bool click;

  InputKeyBase() : time_off(0), time_on(0), key_count(0), flow(false), click(false) { }

  // this should be called from within the interrupt handler
  inline void Trigger()
  {
    click = true;
  }

  bool Read(bool on)
  {
    if ( click )
      {
        on = true;
        click = false;
      }

    uint32_t ct = timeMS();
    bool count = on && time_off != 0xffffffff && (ct - time_off) > KEY_PRESS_DELAY;

    if ( on )
        time_off = 0xffffffff;
    else
        time_off = ct;

    if ( count )
      {
        key_count++;
        time_on = ct;
        flow = false;
      }

    if ( on )
      {
        if ( !flow && ct - time_on > KEY_HOLD_DELAY )
          {
            key_count++;
            flow = true;
            time_on += KEY_HOLD_DELAY;
          }
        else if ( flow && ct - time_on > KEY_REPEAT_DELAY )
          {
            key_count++;
            time_on += KEY_REPEAT_DELAY;
          }
      }

    if ( key_count )
      {
        key_count--;
        return true;
      }

    return false;
  }
};

template <class PIN>
struct InputKey : private InputKeyBase
{
public:
  inline void Trigger() { InputKeyBase::Trigger(); }
  inline bool Read() { return InputKeyBase::Read(PIN::IsOn()); }
};

static InputKey<B_UP> UP_KEY;
static InputKey<B_DOWN> DOWN_KEY;

ISR(INT0_vect)
{
  UP_KEY.Trigger ();
}

ISR(INT1_vect)
{
  DOWN_KEY.Trigger ();
}

static volatile bool reset = false;
// reset button was pressed - clear timing information
ISR(PCINT0_vect)
{
  if ( B_RESET::IsOn() )
      reset = true;
}

// counter storage address in the EEPROM
uint16_t EEMEM eeprom_counter;

static uint8_t old_led_value = 0xff;

void setLeds (uint8_t value)
{
  if ( old_led_value == value )
      return;
  if ( value == 0xff )
    {
      D0::Clear();
      D1::Clear();
    }
  else
    {
      D0::Set(value / 10);
      D1::Set(value-(value/10)*10);
    }
  old_led_value = value;
}

int main()
{
  cli();

  BEEP::Init();
  D0::Init();
  D1::Init();

  XAVR::Pins<B_RESET, B_UP, B_DOWN>::Init();

  // setting up the global_time counter
  TCCR1A = 0;
  TCCR1B = 0x04; // divide by 256 - 32768/256=128 increments per second, 2^16/128=512 seconds for a rollover
  TIMSK1 |= 0x01;
  TCNT1 = 0;

  // up and down buttons
  EICRA = _BV(ISC00) | _BV(ISC01) | _BV(ISC10) | _BV(ISC11);
  EIMSK = 3;

  // reset button
  PCICR = 1;
  PCMSK0 = 1;

  // power control register, disable all possible systems not used by us
  // no point in idling really. compared to what current the leds draw,
  // the mcu is almost nothing
  PRR |= _BV(PRTWI) | _BV(PRTIM0) | _BV(PRSPI) | _BV(PRUSART0) | _BV(PRADC);

  uint32_t countdown = 600;
  uint32_t last_cc = 0xffffffff;
  uint32_t start_time = timeMS();

  // read old timer value from the EEPROM
  countdown = eeprom_read_word (&eeprom_counter);
  if ( countdown > 99 * 60 )
      countdown = 600;
  uint32_t old_countdown = countdown;

  sei();

  while(1)
    {
      uint32_t ct = timeMS();

      if ( UP_KEY.Read() )
        {
          if ( countdown < 60 )
              countdown++;
          else if ( countdown < 99 * 60 )
              countdown += 60;
          last_cc = ct;
          start_time = ct;
        }

      if ( DOWN_KEY.Read() )
        {
          if ( countdown > 60 )
              countdown -= 60;
          else if ( countdown > 0 )
              countdown--;
          last_cc = ct;
          start_time = ct;
        }

      if ( reset )
        {
          start_time = ct;
          reset = false;
        }

      uint32_t time_elapsed = ct - start_time;

      bool blink = false;
      bool beep = false;
      uint16_t v;
      if ( last_cc != 0xffffffff && ct - last_cc < 3500 )
        {
          if ( countdown >= 60 )
              v = countdown / 60;
          else
              v = countdown;
          if ( ct - last_cc > 1000 )
              blink = true;
        }
      else if ( last_cc != 0xffffffff )
        {
          last_cc = 0xffffffff;
          reset = true;
          continue;
        }
      else
        {
          // update the new countdown value to the EEPROM
          if ( countdown != old_countdown )
            {
              cli();
              eeprom_write_word (&eeprom_counter, countdown);
              sei();
              old_countdown = countdown;
            }

          uint32_t ttd; // time to display
          if ( time_elapsed >= countdown * 1000 )
            {
              beep = true;
              blink = true;
              if ( countdown >= 60 )
                  ttd = 1000 * (countdown / 60);
              else
                  ttd = 1000 * countdown;
              if ( time_elapsed > (uint32_t)99 * 60 * 1000 )
                {
                  beep = false;
                  start_time = ct - 1000 * 100 * 60; // to prevent rollovers
                }
            }
          else
            {
              ttd = countdown * 1000 - time_elapsed;
            }

          if ( ttd >= 60000 )
            {
              v = ttd / 60000;
              if ( ttd != (uint32_t)v * 60000 )
                  v++;
            }
          else if ( ttd >= 59000 )
            {
              v = 1;
            }
          else
            {
              v = ttd / 1000;
              if ( ttd != (uint32_t)v * 1000 )
                  v++;
            }
        }

      if ( blink & TCNT1 % 64 < 32 )
        {
          setLeds (0xff);
          BEEP::Clear();
        }
      else
        {
          setLeds (v);

          if ( beep )
              BEEP::Set();
          else
              BEEP::Clear();
        }
    }

  return 0;
}

